# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from os.path import isfile, abspath

from dbmanagr.sources.source import Source

logger = logging.getLogger(__name__)


class AnyPassSource(Source):
    def __init__(self, driver, file_, con_creator):
        Source.__init__(self)
        self.driver = driver
        self.file = file_
        self.con_creator = con_creator

    def list(self):
        if not isfile(self.file):
            from os.path import realpath
            logger.info(
                'File %r does not exist (this file: %r)',
                self.file, realpath(__file__))

            return self._connections

        if not self._connections:
            with open(self.file) as f:
                anypass = f.readlines()

            for line in anypass:
                connection = self.con_creator(
                    self.driver, *line.strip().split(':'))
                self._connections.append(connection)

        logger.debug('Connections: %s', self._connections)

        return self._connections


class AnyFilePassSource(AnyPassSource):
    def list(self):
        if not isfile(self.file):
            from os.path import realpath
            logger.info(
                'File %r does not exist (this file: %r)',
                self.file, realpath(__file__))

            return self._connections

        if not self._connections:
            with open(self.file) as f:
                anypass = f.readlines()

            for line in anypass:
                filepath = abspath(line.strip())
                if isfile(filepath):
                    connection = self.con_creator(
                        self.driver, None, None, filepath, None, None)
                    self._connections.append(connection)

        logger.debug('Connections: %s', self._connections)

        return self._connections



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import sys

from dbmanagr import KIND_VALUE, KIND_FOREIGN_KEY, KIND_FOREIGN_VALUE
from dbmanagr import OPTION_URI_SINGLE_ROW_FORMAT
from dbmanagr import OPTION_URI_MULTIPLE_ROWS_FORMAT
from dbmanagr.wrapper import Wrapper
from dbmanagr.utils import tostring, foreign_key_or_column, find_connection
from dbmanagr.logger import LogWith
from dbmanagr.querybuilder import QueryBuilder, SimplifyMapper
from dbmanagr.exception import UnknownTableException
from dbmanagr.comment import create_comment
from dbmanagr.config import Config
from dbmanagr.writer import Writer
from dbmanagr.sources.source import Source
from dbmanagr.model.value import Value
from dbmanagr.model.row import Row
from dbmanagr.model.tablecomment import COMMENT_TITLE

from .dto import to_dto
from .args import parser
from .writer import SimplifiedWriter

logger = logging.getLogger(__name__)

IMAGE_CONNECTION = 'images/connection.png'
IMAGE_DATABASE = 'images/database.png'
IMAGE_TABLE = 'images/table.png'


def forward_references(row, table, keys, aliases):
    foreign_keys = table.foreign_keys()
    alias = aliases.get(table.name, table.name)

    refs = []
    for key in keys:
        k = key.replace('{0}_'.format(alias), '', 1)
        if key in foreign_keys:
            # Key is a foreign key column
            fk = foreign_keys[key]
            autocomplete = fk.b.table.autocomplete_(
                fk.b.name, row[tostring(key)])
        elif table.column(k).primary_key:
            # Key is a simple column, but primary key
            autocomplete = table.autocomplete_(
                k, row[tostring(key)], OPTION_URI_SINGLE_ROW_FORMAT)
        else:
            # Key is a simple column
            autocomplete = table.autocomplete_(
                k, row[tostring(key)], OPTION_URI_MULTIPLE_ROWS_FORMAT)
        f = foreign_key_or_column(table, k)
        kind = KIND_VALUE
        if f.__class__.__name__ == 'ForeignKey':
            kind = KIND_FOREIGN_KEY
        refs.append(Value(
            row[tostring(key)],
            str(f),
            autocomplete,
            True,
            kind))

    return refs


def back_references(row, table, aliases):
    foreign_keys = table.foreign_keys()

    refs = []
    for key in sorted(
            foreign_keys,
            key=lambda k: foreign_keys[k].a.table.name):
        fk = foreign_keys[key]
        if fk.b.table.name == table.name:
            autocomplete = fk.a.table.autocomplete_(
                fk.a.name, row['{0}_{1}'.format(
                    aliases.get(fk.b.table.name, fk.b.table.name), fk.b.name)],
                OPTION_URI_MULTIPLE_ROWS_FORMAT)
            logger.debug(
                'table.name=%s, fk=%s, autocomplete=%s',
                table.name, fk, autocomplete)
            refs.append(
                Value(
                    fk.a,
                    str(foreign_key_or_column(fk.a.table, fk.a.name)),
                    autocomplete,
                    False,
                    KIND_FOREIGN_VALUE))

    return refs


@LogWith(logger)
def create_databases(connection, options):
    databases = connection.databases()
    if options.database:
        databases = filter(lambda db: options.database in db.name, databases)

    return sorted(databases, key=lambda db: db.name.lower())


@LogWith(logger)
def create_values(connection, table, filter_):
    """Creates row values according to the given filter"""

    builder = QueryBuilder(
        connection,
        table,
        filter_=filter_.filter,
        limit=1,
        simplify=filter_.simplify)

    mapper = None
    keys = None
    if filter_.simplify:
        comment = create_comment(
            table,
            connection.comment(table.name),
            builder.counter,
            builder.aliases,
            None)

        keys = comment.display

        mapper = SimplifyMapper(
            table,
            comment=comment)

    result = connection.queryone(
        builder.build(),
        mapper)

    row = Row(table, result)

    logger.debug('Keys: %s', keys)
    if keys is None:
        keys = sorted(
            row.row.keys(),
            key=lambda key: '' if key == COMMENT_TITLE else tostring(key))
    logger.debug('Keys: %s', keys)

    values = forward_references(row, table, keys, builder.aliases)
    values += back_references(row, table, builder.aliases)

    return values


@LogWith(logger)
def create_tables(connection, options):
    tables = map(lambda (k, t): t, connection.tables().iteritems())
    if options.table:
        tables = filter(
            lambda t: t.name.startswith(options.table),
            tables)

    return sorted(tables, key=lambda t: t.name.lower())


def filter_complete(f):
    if not f:
        raise Exception("No filter given")
    return len(f) > 0 and f.last().rhs is not None


@LogWith(logger)
def create_rows(connection, table, options):
    return sorted(
        connection.rows(
            table,
            options.filter,
            limit=options.limit,
            simplify=options.simplify),
        key=lambda r: r[0])


@LogWith(logger)
def create_columns(table, options):
    return sorted(
        table.columns(options.filter.last().lhs),
        key=lambda c: c.name.lower())


def create(connection, options):
    if options.show == 'connections':
        # print this connection
        return [connection]

    try:
        connection.connect(options.database)

        if options.show == 'databases':
            return create_databases(connection, options)

        if options.show == 'tables':
            return create_tables(connection, options)

        tables = connection.tables()
        if options.table not in tables:
            raise UnknownTableException(options.table, tables.keys())

        table = tables[options.table]
        if options.show == 'columns':
            if filter_complete(options.filter):
                return create_rows(connection, table, options)
            return create_columns(table, options)

        if options.show == 'values':
            return create_values(connection, table, options)
    finally:
        connection.close()


class DatabaseNavigator(Wrapper):
    """The main class"""
    def __init__(self, options):
        Wrapper.__init__(self, options)

        if options.formatter:
            Writer.set(options.formatter())
        else:
            Writer.set(SimplifiedWriter())

    @LogWith(logger, log_result=False)
    def execute(self):
        """The main method that splits the arguments and starts the magic"""
        options = self.options

        return to_dto(self.build(options))

    def build(self, options):
        cons = Source.connections()

        # search exact match of connection
        connection, opts = find_connection(
            cons,
            options,
            lambda con, opts: (
                opts.show != 'connections' and con.matches(opts)))

        if connection is not None:
            return create(connection, opts)

        # print all connections
        return sorted(
            [c for c in cons if c.filter_(options)],
            key=lambda c: c.title().lower())


def execute(args):
    """
    Directly calls the execute method and avoids using the wrapper
    """
    return DatabaseNavigator(Config.init(args, parser)).execute()


def run(args):
    return DatabaseNavigator(Config.init(args, parser)).run()


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    return DatabaseNavigator(Config.init(args, parser)).write()



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.dto import Dto
from dbmanagr.jsonable import from_json
from dbmanagr.utils import filter_keys


class Row(Dto):
    def __init__(
            self,
            table=None,
            row=None,
            autocomplete=None):
        Dto.__init__(self, autocomplete)

        self.table = table
        self.row = row

    def __getitem__(self, i):
        if i is None:
            return None
        if type(i) == unicode:
            i = i.encode('ascii')
        if type(i) is str:
            try:
                return self.row.__dict__[i]
            except BaseException:
                return None
        return self.row[i]

    @staticmethod
    def from_json(d):
        return Row(
            **from_json(filter_keys(d, 'table', 'row', 'autocomplete'))
        )



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import re
import pkgutil
import os
import sys
import uuid

from sqlalchemy.sql.sqltypes import Boolean, Integer, Float, String, \
    Date, Time, DateTime, _Binary

# from dbmanagr.logger import LogWith

NAMES = [
    'name', 'title', 'key', 'text', 'first_name', 'username', 'user_name',
    'last_name', 'email', 'comment', 'street', 'city'
]
NAME_SUFFIXES = ['name', 'title', 'key', 'text']

logger = logging.getLogger(__name__)


def hash_(s):
    try:
        return str(uuid.uuid3(uuid.NAMESPACE_DNS, s.encode('ascii', 'ignore')))
    except:
        return hash(s)


def module_installed(*modules):
    installed = map(lambda m: m[1], pkgutil.iter_modules())
    for module in modules:
        if module in installed:
            return module
    return None


def prefix(item, separator='\\.'):
    return re.sub('([^{0}]*){0}.*'.format(separator), '\\1', item)


def prefixes(items):
    if items is None:
        return None

    if type(items) is dict:
        return dict(filter(
            lambda (k, v): len(prefix(k)), map(
                lambda (k, v): (prefix(k), v),
                filter(
                    lambda (k, v): '.' not in k,
                    items.iteritems())
                )
        ))

    return set(filter(len, map(prefix, items)))


def remove_prefix(prefix, items):
    if items is None:
        return None

    p = '%s.' % prefix

    if type(items) is dict:
        return dict(map(
            lambda (k, v): (re.sub('^%s' % p, '', k), v),
            filter(
                lambda (k, v): k.startswith(p),
                items.iteritems())
        ))

    return [re.sub('^%s' % p, '', i) for i in items if i.startswith(p)]


def tostring(key):
    if isinstance(key, unicode):
        return key.encode('ascii', errors='ignore')
    return key


def dictsplus(dicts, key, value):
    for d in dicts:
        dictplus(d, key, value)
    return dicts


def dictplus(d, key, value):
    d[key] = value
    return d


def dictminus(d, *keys):
    r = dict(d)
    for key in keys:
        if key in d:
            del r[key]
    return r


def getorelse(i, e):
    if i is not None:
        return i
    return e


def is_name(name):
    return lambda c: c.name == name


def ends_with(suffix):
    return lambda c: c.name.endswith(suffix)


# @LogWith(logger)
def create_title(comment, columns, exclude=None):
    if exclude is None:
        exclude = []

    # Find certain column names (but their type is not an integer - integers
    # are no good names)
    for name in filter(lambda n: n not in exclude, NAMES):
        for c in filter(is_name(name), columns):
            if not isinstance(c.type, Integer):
                return (name, '{%s}' % c.name)

    # Find first column that ends with any of certain suffixes
    for suffix in filter(lambda n: n not in exclude, NAME_SUFFIXES):
        for c in filter(ends_with(suffix), columns):
            if not isinstance(c.type, Integer):
                return (c.name, '{%s}' % c.name)

    # Use the comment id, if any
    if comment and comment.id:
        return ('id', '{%s}' % comment.id)

    # Default: use the first column
    if len(columns) > 0:
        return (columns[0].name, '{%s}' % columns[0].name)

    return (None, None)


# @LogWith(logger)
def foreign_key_or_column(table, column):
    fk = table.foreign_key(column)
    if fk:
        return fk
    return table.column(column)


def operation(column, operator, value):
    from dbmanagr import OPERATORS
    return OPERATORS.get(operator)(column, value)


def unicode_decode(arg):
    if type(arg) is list:
        return map(unicode_decode, arg)
    if type(arg) is unicode:
        return arg
    if type(arg) in [int, bool, float]:
        return unicode(arg)
    return arg.decode('utf-8')


def mute_stderr(f):
    def wrapper(*args, **kwargs):
        devnull = open(os.devnull, 'w')
        stderr, sys.stderr = sys.stderr, devnull
        try:
            return f(*args, **kwargs)
        finally:
            sys.stderr.close()
            sys.stderr = stderr

    return wrapper


def matches(name, patterns):
    for pattern in patterns:
        p = re.compile(replace_wildcards(pattern))
        if p.match(name):
            return True

    return False


def replace_wildcards(pattern):
    if not pattern:
        return pattern

    return pattern.replace('*', '.*')


def primary_key_or_first_column(table):
    column = table.primary_key
    if not column:
        column = table.column(0).name
    return column


def filter_keys(d, *keys):
    return dict(filter(lambda (k, v): k in keys, d.iteritems()))


def freeze(d):
    if isinstance(d, dict):
        return frozenset((key, freeze(value)) for key, value in d.items())
    elif isinstance(d, list):
        return tuple(freeze(value) for value in d)
    return d


def escape_statement(stmt):
    """Escapes the given statement for use with SQLAlchemy"""

    return stmt.replace('%', '%%')


def find_connection(cons, options, matcher):
    logger.debug(
        'find_connection(cons=%s, options=%s, matcher=%s)',
        cons, options, matcher)
    for con in cons:
        opts = options.get(con.dbms)
        logger.debug('matcher(con=%s, opts=%s)', con, opts)
        if matcher(con, opts):
            return (con, opts)

    return (None, None)


def to_dict(opts, d=None):
    """
    Creates a dictionary from the given list of strings. The dict keys are
    separated by dots. This can be used to create the includes/excludes/
    substitutes given as command line arguments.
    """

    if d is None:
        d = {}
    if opts is None:
        return d

    for opt in opts:
        s = opt.split('=', 1)
        key = s[0]
        val = None
        if len(s) > 1:
            val = s[1]

        if key == '':
            if d == {}:
                # Overwrite newly created dict
                d = None
            return d

        if '.' in key:
            # Recurse into keys
            pfx = prefix(key)
            if pfx not in d or type(d[pfx]) is bool:
                d[pfx] = {}
            d[pfx] = to_dict(remove_prefix(pfx, [opt]), d[pfx])
        else:
            d[prefix(key)] = val

    return d


# Example
#
# a:
#   b: False
#   c: True
# d:
# e:
#   f:
#     g:
#
# When Included: a, a.c, d, e, e.f, e.f.g
# When Excluded: a.c, d, e.f.g
#
def is_included(name, d):
    """Checks the given content selection dict for inclusion of name"""

    if d is None:
        # None means included
        return True
    if d is False:
        return False
    # Type must be dict, then any value (None, dict, True) except for False
    # includes name
    if type(d) is dict:
        if '*' in d:
            # Wildcard matches any element within this level
            return True
        return d.get(name, False) is not False
    return False


def is_excluded(name, d):
    """Checks the given content selection dict for exclusion of name"""

    if d is None:
        # None means excluded
        return True
    if d is False:
        return False
    if type(d) is dict:
        if '*' in d:
            # Wildcard matches any element within this level
            return True
        val = d.get(name, True)
        return val is None or val is False
    return True


def selection(name, d):
    """Retrieves the content selection for the given name"""

    if type(d) is dict:
        return d.get(name, False)
    return False


def is_node(tree, name):
    """Checks whether the name is part of the tree dict"""

    return (
        type(tree) is dict
        and name in tree
        and type(tree[name]) is dict)


def to_ref(parent, key):
    if parent is None:
        return key
    return '{0}.{1}'.format(parent, key)


def to_forward_ref(ref):
    if ref.endswith('*'):
        return ref
    return '{0}.'.format(ref)


def to_yaml_type(type_):
    if isinstance(type_, Integer):
        return 'int'
    if isinstance(type_, Float):
        return 'float'
    if isinstance(type_, String):
        return 'str'
    if isinstance(type_, Boolean):
        return 'bool'
    if (isinstance(type_, DateTime)
            or isinstance(type_, Date)
            or isinstance(type_, Time)):
        return 'str'
    if isinstance(type_, _Binary):
        return 'binary'
    return 'str'


def shell_escape(s):
    if type(s) in (str, unicode):
        if "'" in s:
            return u'"{0}"'.format(s)
        return u"'{0}'".format(s)
    return unicode(s)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import logging

from dbmanagr.logger import logger
from dbmanagr.options import Options
from dbmanagr.utils import unicode_decode

# load sources
from importlib import import_module
import dbmanagr.driver as driver
for mod in driver.__all__:
    import_module('dbmanagr.driver.{0}'.format(mod)).init()


class Config(object):
    @staticmethod
    def init(argv, parser):
        options = Options(unicode_decode(argv), parser)

        logging.basicConfig(
            stream=options.logfile,
            level=options.loglevel,
            format="%(asctime)s %(levelname)s %(name)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S")

        logger.info(
            """
###
### %s called with args: %s
###""",
            parser.prog, options.argv)

        logger.debug("Options: %s", options)
        logger.debug(
            "Environment: %s",
            dict(filter(
                lambda (k, v): k.startswith('DBNAV_'),
                os.environ.iteritems())))

        return options



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.writer import FormatWriter
from dbmanagr.formatter import Formatter, DefaultFormatter


def value_from_column(column, config):
    if config.verbose > 3:
        return '{0}?{1}'.format(
            column.autocomplete().split('?')[0], column.ddl())
    if config.verbose > 2:
        return column.autocomplete()
    if config.verbose > 1:
        return '/'.join(column.autocomplete().split('/')[1:])
    if config.verbose > 0:
        return unicode(column)
    return column_name(column, config)


def column_name(column, config):
    if config.compare_ddl:
        return column.ddl()
    return column.name


class DiffWriter(FormatWriter):
    def __init__(self, left=None, right=None):
        FormatWriter.__init__(self, u'{0}\n')
        Formatter.set(DefaultFormatter())
        self.left = left
        self.right = right

    def str(self, items):
        items = [i for i in items]  # resolves generator
        if not items:
            return self.items_format.format('No differences found')
        s = self.item_separator.join(
            map(self.itemtostring, self.filter_(items)))
        return self.items_format.format(s)

    def itemtostring(self, item):
        """We receive a tuple (left, right), for which any part may be empty"""
        left, right = item
        a = []
        if left:
            a.append(u'< {0}'.format(value_from_column(left, self.left)))
        if right:
            a.append(u'> {0}'.format(value_from_column(right, self.right)))
        return u'\n'.join(a)


class DiffColumnWriter(DiffWriter):
    def __init__(self, left=None, right=None):
        DiffWriter.__init__(self, left, right)
        self.left = left
        self.right = right

    def itemtostring(self, item):
        """We receive a tuple (left, right), for which any part may be empty"""
        left, right = item
        s = ''
        if left:
            val = value_from_column(left, self.left)
            s += u'{1}{0}'.format(u' ' * (42 - len(val)), val)
            if right:
                s += ' | '
        else:
            s += u'{0} > '.format(u' ' * 42)
        if right:
            val = value_from_column(right, self.right)
            s += val
        else:
            s += ' <'
        return s


class DiffTestWriter(DiffWriter):
    pass



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from dbmanagr.logger import LogWith
from dbmanagr.options import restriction, FileOptionsParser
from dbmanagr.driver import DatabaseDriver

logger = logging.getLogger(__name__)


class SQLiteDriver(DatabaseDriver):
    @LogWith(logger)
    def restriction(self, *args):
        return restriction(*args)

    def statement_activity(self, con):
        return []

    def __repr__(self):
        return str(self.__dict__)


class SQLiteOptionsParser(FileOptionsParser):
    def create_driver(self):
        return SQLiteDriver()



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import datetime
import sqlalchemy
import re

from decimal import Decimal

from dbmanagr.exception import BusinessException


def to_key(key):
    return re.sub(r"_+$", '', key)


def as_json(obj):
    if isinstance(obj, (int, long, float, bool)) or obj is None:
        return obj
    if isinstance(obj, dict):
        return dict(map(
            lambda (k, v): (to_key(k), as_json(v)),
            obj.iteritems())
        )
    if isinstance(obj, (datetime.datetime, datetime.date)):
        return {
            '__cls__': 'datetime.{}'.format(obj.__class__.__name__),
            'value': obj.isoformat()
        }
    if isinstance(obj, sqlalchemy.util.KeyedTuple):
        d = {
            '__cls__': 'sqlalchemy.util.KeyedTuple'
        }
        d.update(dict(map(as_json, obj.__dict__.iteritems())))
        return d
    if isinstance(obj, sqlalchemy.sql.sqltypes.NullType):
        return None
    if isinstance(obj, (tuple, list, set)):
        return map(as_json, obj)
    if isinstance(obj, Jsonable):
        d = {
            '__cls__': str(obj.__class__)
        }
        for key, value in obj.__dict__.iteritems():
            if not key.startswith('_'):
                if isinstance(value, Jsonable):
                    d[to_key(key)] = value.as_json()
                else:
                    d[to_key(key)] = as_json(value)
        return d
    return unicode(obj)


def import_class(name):
    parts = name.split('.')
    if len(parts) > 1:
        mod = __import__('.'.join(parts[:-1]), fromlist=[str(parts[-1])])
        return getattr(mod, parts[-1])
    return None


def from_json(d):
    if type(d) is dict:
        if '__cls__' in d:
            classname = d['__cls__']
            if (classname.endswith('Exception')
                    or classname.endswith('Error')
                    or classname.endswith('Exit')):
                return BusinessException(d['message'])
            if classname == 'sqlalchemy.util.KeyedTuple':
                from sqlalchemy.util import KeyedTuple
                return KeyedTuple(
                    map(lambda k: from_json(d[k]), d['_labels']),
                    labels=d['_labels'])
            if classname == 'datetime.datetime':
                from datetime import datetime
                try:
                    return datetime.strptime(d['value'], "%Y-%m-%dT%H:%M:%S")
                except ValueError:
                    return datetime.strptime(
                        d['value'], "%Y-%m-%dT%H:%M:%S.%f")
            if classname == 'datetime.date':
                from datetime import datetime
                return datetime.strptime(d['value'], "%Y-%m-%d").date()
            cls = import_class(classname)
            if hasattr(cls, 'from_json'):
                return cls.from_json(d)
        return dict(map(lambda (k, v): (k, from_json(v)), d.iteritems()))
    if type(d) is list or type(d) is tuple:
        return map(from_json, d)
    if type(d) is Decimal:
        if d % 1 == 0:
            return int(d)
        else:
            return float(d)
    return d


class Jsonable(object):
    def as_json(self):
        return as_json(self)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import argparse
import sys

from dbmanagr.args import parent_parser, format_group, create_parser

from .writer import ExecuteWriter, SqlInsertWriter, ExecuteTestWriter

parent = parent_parser()

group = format_group(parent, ExecuteTestWriter)
group.add_argument(
    '-D',
    '--default',
    help='output format: tuples',
    dest='formatter',
    action='store_const',
    const=ExecuteWriter)
group.add_argument(
    '-I',
    '--insert',
    help='output format: SQL insert statements',
    dest='formatter',
    action='store_const',
    const=SqlInsertWriter)

parser = create_parser(
    prog='dbexec',
    description='Executes the SQL statements from the given file on the '
                'database specified by the given URI',
    parents=[parent])
parser.add_argument(
    'uri',
    help='the URI to parse (format for PostgreSQL/MySQL: user@host/database; '
         'for SQLite: databasefile.db)')
parser.add_argument(
    'infile',
    default=[sys.stdin],
    help='the path to the file containing the SQL query to execute (default: '
         'standard input)',
    type=argparse.FileType('r'),
    nargs='*')
parser.add_argument(
    '-s',
    '--statements',
    help='the statements to execute (infile will be ignored when this '
         'parameter is given)')
parser.add_argument(
    '-p',
    '--progress',
    default=-1,
    type=int,
    help='show progress after this amount of executions when inserting/'
         'updating large data sets (default: %(default)s)')
parser.add_argument(
    '-t',
    '--table-name',
    default='__TABLE__',
    help='the table name for generic select statements (default: %(default)s)')
parser.add_argument(
    '-n',
    '--dry-run',
    action='store_true',
    default=False,
    help='Do a rollback after execution')
parser.add_argument(
    '--isolate-statements',
    action='store_true',
    default=False,
    help='Wrap each statement in a separate transaction - this allows '
         'continuing execution even if an SQL statement fails')
parser.add_argument(
    '--mute-errors',
    action='store_true',
    default=False,
    help='Don\'t display error message for failing statements')
parser.add_argument(
    '-v',
    '--verbose',
    action='count',
    help='specify the verbosity of the output, increase the number of '
         'occurences of this option to increase verbosity')



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.jsonable import Jsonable
from dbmanagr.formatter import Formatter

PRIMARY_KEY_OPTIONS = {
    True: '*',
    False: ''
}
NULLABLE_OPTIONS = {
    True: '?',
    False: '',
    None: ''
}


class BaseNode(Jsonable):
    def __eq__(self, o):
        return hash(self) == hash(o)


class ColumnNode(BaseNode):
    def __init__(self, column, indent=0):
        self.column = column
        self.indent = indent

    def __hash__(self):
        return hash(str(self.column))


class ForeignKeyNode(BaseNode):
    def __init__(self, fk, parent=None, indent=0):
        self.fk = fk
        self.parent = parent
        self.indent = indent

    def __getattr__(self, name):
        return getattr(self.fk, name)

    def __hash__(self):
        return hash(str(self.fk))


class TableNode(BaseNode):
    def __init__(self, table, include=None, exclude=None, indent=0):
        if include is None:
            include = {}
        if exclude is None:
            exclude = {}

        self.table = table
        self.include = include
        self.exclude = exclude
        self.indent = indent

    def __hash__(self):
        return hash(self.table.autocomplete())


class NameNode(BaseNode):
    def __init__(self, name, indent=0):
        self.name = name
        self.indent = indent

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return '{0}{1}'.format('  ' * self.indent, self.name)

    def format(self):
        return Formatter.format_name_node(self)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.dto import Dto
from dbmanagr.jsonable import from_json


class ForeignKey(Dto):
    def __init__(self, a=None, b=None):
        Dto.__init__(self)

        self.a = a
        self.b = b

    def __str__(self):
        return '{a} -> {b}'.format(a=self.a, b=self.b)

    @staticmethod
    def from_json(d):
        return ForeignKey(
            a=from_json(d.get('a')),
            b=from_json(d.get('b'))
        )



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from string import capwords
from decimal import Decimal

from dbmanagr.utils import matches
from dbmanagr.writer import FormatWriter
from dbmanagr.formatter import Formatter, DefaultFormatter


def included(name, include, exclude):
    """Per default all columns are included"""

    if matches(name, include):
        return True

    if matches(name, exclude):
        return False

    return True


class SqlInsertWriter(FormatWriter):
    def __init__(self, options):
        FormatWriter.__init__(
            self,
            u'{0}\n',
            u'insert into {table} ({columns}) values ({values});')
        Formatter.set(DefaultFormatter())
        self.options = options

    def itemtostring(self, item):
        row = item.row
        exclude = item.exclude
        include = item.include
        substitutes = item.substitutes
        return self.item_format.format(
            table=self.options.escape_keyword(row.table.name),
            columns=self.create_columns(row, include, exclude),
            values=self.create_values(row, include, exclude, substitutes))

    def create_columns(self, row, include, exclude):
        return u','.join(
            map(lambda col: self.options.escape_keyword(col.name),
                filter(
                    lambda col: included(col.name, include, exclude),
                    row.table.columns)))

    def create_values(self, row, include, exclude, substitutes):
        return u','.join(
            map(lambda col: substitutes.get(
                col.name,
                self.options.format_value(col, row[col.name])),
                filter(
                    lambda col: included(col.name, include, exclude),
                    row.table.columns)))


class SqlUpdateWriter(FormatWriter):
    def __init__(self, options):
        FormatWriter.__init__(
            self,
            u'{0}\n',
            u'update {table} set {values} where {restriction};')
        Formatter.set(DefaultFormatter())
        self.options = options

    def itemtostring(self, item):
        row = item.row
        include = item.include
        exclude = item.exclude
        # substitutes = item.substitutes
        table = row.table
        return self.item_format.format(
            table=self.options.escape_keyword(table.name),
            values=self.create_values(row, include, exclude),
            restriction=self.create_restriction(
                row,
                filter(lambda col: col.primary_key, table.columns)))

    def create_values(self, row, include, exclude):
        # Substitutes do not make sense here, or do they?
        return u', '.join(map(
            lambda col: self.options.restriction(
                None, col, '=', row[col.name], map_null_operator=False),
            filter(
                lambda col: (
                    not col.primary_key
                    and included(col.name, include, exclude)),
                row.table.columns)))

    def create_restriction(self, row, pks):
        return u' and '.join(map(
            lambda col: self.options.restriction(
                None, col, '=', row[col.name]),
            pks))


class SqlDeleteWriter(FormatWriter):
    def __init__(self, options):
        FormatWriter.__init__(
            self,
            u'{0}\n',
            u'delete from {table} where {restriction};')
        Formatter.set(DefaultFormatter())
        self.options = options

    def prepare(self, items):
        # Fixes issue #4
        items = [i for i in items]  # resolves generator
        items.reverse()
        return items

    def itemtostring(self, item):
        row = item.row
        table = row.table
        return self.item_format.format(
            table=self.options.escape_keyword(table.name),
            restriction=self.create_restriction(
                row, filter(lambda col: col.primary_key, table.columns)))

    def create_restriction(self, row, pks):
        return u' and '.join(
            map(
                lambda col: self.options.restriction(
                    None, col, '=', row[col.name]),
                pks))


def yaml_format_entity(name):
    return capwords(name, '_').replace('_', '')


def yaml_format_field(name):
    s = yaml_format_entity(name)
    return s[:1].lower() + s[1:]


def yaml_field(col, table):
    if table and col in table.foreign_keys:
        fk = table.foreign_keys[col]
        return yaml_format_field(fk.b.tablename)
    return yaml_format_field(col)


def yaml_value(col, table, value):
    if table and col in table.foreign_keys:
        fk = table.foreign_keys[col]
        return u'*{table}_{id}'.format(
            table=fk.b.tablename.replace('_', ''),
            id=value)
    if value is None:
        return u'!!null null'
    if type(value) is float:
        return u'!!float %f' % value
    if type(value) is int:
        return u'!!int %d' % value
    if type(value) is bool:
        return {True: u'Yes', False: u'No'}.get(value, u'No')
    if type(value) is Decimal:
        if value % 1 == 0:
            return u'!!int %d' % value
        else:
            return u'!!float %f' % value
    if type(value) in (str, unicode):
        return u'!!str %s' % value
    return value


class YamlWriter(FormatWriter):
    def __init__(self, options=None):
        FormatWriter.__init__(
            self,
            u'{0}\n',
            u"""{prefix}    - &{table}_{id} !!{package}.{model}
        {tuples}""")
        Formatter.set(DefaultFormatter())
        self.package = options.package
        self.last_table = None

    def itemtostring(self, item):
        row = item.row
        exclude = item.exclude
        substitutes = item.substitutes
        table = row.table
        tablename = table.name.replace('_', '')
        prefix = ''
        if self.last_table != table:
            if self.last_table:
                prefix = u"\n{0}s:\n".format(tablename)
            else:
                prefix = u"{0}s:\n".format(tablename)
            self.last_table = table
        return self.item_format.format(
            table=tablename,
            id=row[0],
            prefix=prefix,
            package=self.package,
            model=yaml_format_entity(table.name),
            tuples=self.create_tuples(row, exclude, substitutes))

    def create_tuples(self, row, exclude, substitutes):
        return u"\n        ".join(map(
            lambda col: u'{0}: {1}'.format(
                yaml_field(col.name, row.table),
                substitutes.get(
                    col.name,
                    yaml_value(col.name, row.table, row[col.name]))),
            filter(lambda col: col.name not in exclude, row.table.columns)))


class FormattedWriter(FormatWriter):
    def __init__(self, options=None):
        FormatWriter.__init__(
            self,
            u'{0}\n',
            options.format)
        Formatter.set(DefaultFormatter())

    def itemtostring(self, item):
        d = dict(map(
            lambda col: (col.name, item.row[col.name]),
            item.row.table.columns))
        return self.item_format.format(**d)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from difflib import get_close_matches

from dbmanagr.logger import logger

CONNECTION_NOT_FOUND = 'Connection "{0}" was not found ({1})'
TABLE_NOT_FOUND = 'Table "{0}" was not found ({1})'
COLUMN_NOT_FOUND = 'Column "{0}" was not found on table "{1}" ({2})'
CLOSE_MATCHES = 'close matches: {0}'
NO_CLOSE_MATCHES = 'no close matches in: {0}'


def unknown_connection_message(connection, haystack):
    matches = get_close_matches(connection, haystack)
    if not matches:
        return CONNECTION_NOT_FOUND.format(connection, 'no close matches')
    return CONNECTION_NOT_FOUND.format(
        connection,
        CLOSE_MATCHES.format(u', '.join(matches)))


def unknown_table_message(tablename, haystack):
    if tablename is None:
        tablename = '?'
    matches = get_close_matches(tablename, haystack)
    if not matches:
        return TABLE_NOT_FOUND.format(
            tablename,
            NO_CLOSE_MATCHES.format(u', '.join(haystack)))
    return TABLE_NOT_FOUND.format(
        tablename,
        CLOSE_MATCHES.format(u', '.join(matches)))


def unknown_column_message(table, column, haystack=None):
    if haystack is None:
        haystack = map(lambda c: c.name, table.columns())
    logger.debug('haystack: %s', haystack)
    matches = get_close_matches(column, haystack)
    if not matches:
        return COLUMN_NOT_FOUND.format(
            column,
            table.name if table else '?',
            NO_CLOSE_MATCHES.format(u', '.join(haystack)))
    return COLUMN_NOT_FOUND.format(
        column,
        table.name if table else '?',
        CLOSE_MATCHES.format(u', '.join(matches)))


class BusinessException(Exception):
    pass


class UnknownConnectionException(BusinessException):
    def __init__(self, connection, haystack):
        super(UnknownConnectionException, self).__init__(
            unknown_connection_message(connection, haystack))


class UnknownTableException(BusinessException):
    def __init__(self, tablename, haystack):
        super(UnknownTableException, self).__init__(
            unknown_table_message(tablename, haystack))


class UnknownColumnException(BusinessException):
    def __init__(self, table, column, haystack=None):
        super(UnknownColumnException, self).__init__(
            unknown_column_message(table, column, haystack))



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.dto import Dto
from dbmanagr.jsonable import from_json
from dbmanagr.utils import filter_keys


class Column(Dto):
    def __init__(
            self,
            name=None,
            tablename=None,
            type_=None,
            nullable=None,
            default=None,
            autoincrement=None,
            primary_key=None,
            autocomplete=None):
        Dto.__init__(self, autocomplete)

        self.name = name
        self.tablename = tablename
        self.type = type_
        self.nullable = nullable
        self.default = default
        self.autoincrement = autoincrement
        self.primary_key = primary_key

    def __str__(self):
        if self.tablename:
            return '%s.%s' % (self.tablename, self.name)
        return self.name

    def ddl(self):
        return '{0} {1}{2}{3}'.format(
            self.name,
            self.type,
            {False: ' not null'}.get(self.nullable, ''),
            {None: ''}.get(self.default, ' default {0}'.format(self.default)))

    @staticmethod
    def from_json(d):
        return Column(
            **from_json(
                filter_keys(
                    d,
                    'name', 'tablename', 'type', 'nullable', 'default',
                    'autoincrement', 'primary_key', 'autocomplete'
                )
            )
        )



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os
import logging
import pdb
import urllib2
import json
import ijson

from dbmanagr.writer import Writer
from dbmanagr import logger as log
from dbmanagr.jsonable import from_json

COMMANDS = {
    'dbdiff': 'differ',
    'dbexec': 'executer',
    'dbexport': 'exporter',
    'dbgraph': 'grapher',
    'dbnav': 'navigator'
}


class Wrapper(object):
    def __init__(self, options=None):
        self.options = options

    def write(self):
        try:
            sys.stdout.write(Writer.write(self.run()))
        except BaseException as e:
            log.logger.exception(e)
            return -1
        return 0

    def execute(self):  # pragma: no cover
        """To be overridden by sub classes"""
        pass

    def run(self):
        try:
            if (
                    self.options is not None
                    and self.options.daemon):  # pragma: no cover
                log.logger.debug('Executing remotely')
                return self.executer(*sys.argv)

            log.logger.debug('Executing locally')
            return self.execute()
        except BaseException as e:
            log.logger.exception(e)
            if log.logger.getEffectiveLevel() <= logging.DEBUG:
                # Start post mortem debugging only when debugging is enabled
                if os.getenv('UNITTEST', 'False') == 'True':
                    raise
                if self.options.trace:  # pragma: no cover
                    pdb.post_mortem(sys.exc_info()[2])
            else:
                # Show the error message if log level is INFO or higher
                log.log_error(e)  # pragma: no cover

    def executer(self, *args):  # pragma: no cover
        """Execute remotely"""

        options = self.options

        try:
            # from dbmanagr import daemon
            # if not daemon.is_running(options):
            #     daemon.start_server(options)

            url = 'http://{host}:{port}/{path}'.format(
                host=options.host,
                port=options.port,
                path=COMMANDS[options.prog])
            request = json.dumps(args[1:])

            log.logger.debug('Request to %s:\n%s', url, request)

            response = urllib2.urlopen(url, request)

            for i in ijson.items(response, 'item'):
                yield from_json(i)
        except urllib2.HTTPError as e:
            raise from_json(json.load(e))
        except urllib2.URLError as e:
            log.logger.error('Daemon not available: %s', e)
        except BaseException as e:
            log.logger.exception(e)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.jsonable import Jsonable
from dbmanagr.utils import freeze


class Dto(Jsonable):
    def __init__(self, autocomplete=None):
        self.autocomplete_ = autocomplete

    def autocomplete(self):
        return self.autocomplete_

    def __hash__(self):
        return hash(freeze(self.__dict__))

    def __eq__(self, o):
        return hash(self) == hash(o)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from plistlib import readPlist
from os.path import isfile

from dbmanagr.sources.source import Source

logger = logging.getLogger(__name__)


class NavicatSource(Source):
    def __init__(self, uri, file_, key, con_creator):
        Source.__init__(self)
        self.uri = uri
        self.file = file_
        self.key = key
        self.con_creator = con_creator

    def list(self):
        if not isfile(self.file):
            from os.path import realpath
            logger.warn(
                'File %r does not exist (this file: %r)',
                self.file, realpath(__file__))

            return self._connections

        if not self._connections:
            plist = readPlist(self.file)

            # Note: only works with SQLite ATM - passwords are encrypted within
            # Navicat config files
            for _, v in plist[self.key]['servers'].items():
                connection = self.con_creator(
                    self.uri,
                    None,
                    None,
                    v['dbfilename'],
                    None,
                    None)
                self._connections.append(connection)

        return self._connections



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import codecs
import sys

from dbmanagr.logger import LogWith
from dbmanagr.formatter import Formatter, TestFormatter, DefaultFormatter

UTF8Writer = codecs.getwriter('utf8')
sys.stdout = UTF8Writer(sys.stdout)

logger = logging.getLogger(__name__)


class DefaultWriter(object):
    def write(self, items):
        return self.str(items)

    def str(self, items):
        return map(self.itemtostring, items)

    def itemtostring(self, item):
        return unicode(item)


class StdoutWriter(DefaultWriter):
    def __init__(
            self,
            items_format=u"Title\tSubtitle\tAutocomplete\n{0}\n",
            item_format=u"{title}\t{subtitle}\t{autocomplete}",
            item_separator=u"\n",
            format_error_format=u'{0}'):
        self.items_format = items_format
        self.item_format = item_format
        self.item_separator = item_separator
        self.format_error_format = format_error_format

        Formatter.set(DefaultFormatter())

    def filter_(self, items):
        return items

    def str(self, items):
        items = self.prepare(items)
        items = self.filter_(items)
        s = self.item_separator.join(map(self.itemtostring, items))
        return self.items_format.format(s)

    def prepare(self, items):
        if not items:
            return []
        return items

    def itemtostring(self, item):
        return self.item_format.format(
            item=unicode(item), **item.__dict__)


class FormatWriter(StdoutWriter):
    def __init__(
            self,
            items_format=u"""Title\tSubtitle\tAutocomplete
{0}
""",
            item_format=u"""{title}\t{subtitle}\t{autocomplete}""",
            item_separator=u"""
""",
            format_error_format=u'{0}'):
        StdoutWriter.__init__(
            self,
            items_format,
            item_format,
            item_separator,
            format_error_format)

    def itemtostring(self, item):
        return item.format()


class TestWriter(FormatWriter):
    def __init__(
            self,
            items_format=u"""Title\tAutocomplete
{0}""",
            item_format=u"""{title}\t{autocomplete}"""):
        FormatWriter.__init__(self, items_format, item_format)
        Formatter.set(TestFormatter())


class Writer(object):
    writer = StdoutWriter()

    @staticmethod
    def set(arg):
        Writer.writer = arg

    @staticmethod
    @LogWith(logger, log_args=False, log_result=False)
    def write(items):
        return Writer.writer.write(items)

    @staticmethod
    @LogWith(logger, log_args=False, log_result=False)
    def itemtostring(item):
        return Writer.writer.itemtostring(item)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

__drivers__ = []

KIND_VALUE = 'value'
KIND_FOREIGN_KEY = 'foreign-key'
KIND_FOREIGN_VALUE = 'foreign-value'

IMAGE_VALUE = 'images/value.png'
IMAGE_FOREIGN_KEY = 'images/foreign-key.png'
IMAGE_FOREIGN_VALUE = 'images/foreign-value.png'

OPTION_URI_SINGLE_ROW_FORMAT = u'%s%s/?%s'
OPTION_URI_MULTIPLE_ROWS_FORMAT = u'%s%s?%s'

OPERATORS = {
    '=': lambda c, v: c.__eq__(v),
    '!=': lambda c, v: c.__ne__(v),
    '~': lambda c, v: c.like(v),
    '*': lambda c, v: c.like(v),
    '>': lambda c, v: c.__gt__(v),
    '>=': lambda c, v: c.__ge__(v),
    '<=': lambda c, v: c.__le__(v),
    '<': lambda c, v: c.__lt__(v),
    'in': lambda c, v: c.in_(v),
    ':': lambda c, v: c.in_(v)
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from setuptools import setup, find_packages


def get_version():
    """
    Gets the latest version number out of the package, saving us from
    maintaining it in multiple places.
    """
    local_results = {}
    execfile('src/dbmanagr/version.py', {}, local_results)
    return local_results['__version__']

setup(
    name="dbmanagr",
    version=get_version(),
    description="The database manager for the command line",
    url="https://github.com/resamsel/dbmanagr",
    author="René Samselnig",
    author_email="me@resamsel.com",
    license="GPLv3+",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2 :: Only',
        'Topic :: Database',
        'Topic :: Database :: Front-Ends',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities'
    ],
    keywords="database navigator exporter grapher postgres sqlite graphviz"
             " diff",

    packages=find_packages(
        'src', exclude=['*.tests', '*.tests.*', 'tests.*', 'tests']),
    package_dir={'': 'src'},

    test_suite='tests',

    # dependencies
    install_requires=[
        'sqlalchemy==0.9.10',
        'sqlparse>=0.1.13',
        'ijson>=2.0',
        'pyyaml>=3.11',
        # 'psycopg2>=2.5.1',
        # 'mysql-python>=1.2.5'
    ],

    tests_require=[
        'flake8>=2.2.5',
        'pep8-naming>=0.2.2',
        'flake8-todo>=0.3',
        'nose>=1.3.4',
        'coverage>=3.7.1',
        'pylint>=1.4.3',
        'pg8000>=1.08',
        'pymysql>=0.6.2',
    ],

    entry_points={
        'console_scripts': [
            'dbnav = dbmanagr.command.navigator:main',
            'dbexport = dbmanagr.command.exporter:main',
            'dbgraph = dbmanagr.command.grapher:main',
            'dbexec = dbmanagr.command.executer:main',
            'dbdiff = dbmanagr.command.differ:main',
            'dbdaemon = dbmanagr.command.daemon:main',
            'dbstac = dbmanagr.command.activity:main',
            'dbargs = dbmanagr.command.argumentor:main',
        ]
    },

    data_files=[
#        (
#            '/usr/local/etc/bash_completion.d',
#            ['resources/bash_completion/dbmanagr']
#        )
    ]
)



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = ["databaseconnection", "driver"]

from os.path import expanduser
from os import getenv
from collections import OrderedDict

from dbmanagr import __drivers__
from dbmanagr.utils import module_installed
from dbmanagr.sources.source import Source
from dbmanagr.sources.dbexplorer import DBExplorerSource
from dbmanagr.sources.anypass import AnyFilePassSource
from dbmanagr.sources.navicat import NavicatSource
from dbmanagr.options import Options
from .databaseconnection import SQLiteConnection
from .driver import SQLiteOptionsParser

DRIVERS = OrderedDict([
    ('sqlite3', 'sqlite+pysqlite:///{file}')
])


def init_sqlite(
        uri,
        dbexplorer_config,
        sqlitepass_config,
        navicat_config1,
        navicat_config2=None):
    Source.sources.append(
        DBExplorerSource(uri, dbexplorer_config, 'sqlite', SQLiteConnection))
    Source.sources.append(
        AnyFilePassSource(uri, sqlitepass_config, SQLiteConnection))
    Source.sources.append(
        NavicatSource(uri, navicat_config1, 'SQLite', SQLiteConnection))
    if navicat_config2:
        Source.sources.append(
            NavicatSource(uri, navicat_config2, 'SQLite', SQLiteConnection))


def init():
    module = module_installed(*DRIVERS.keys())
    if not module:
        return

    __drivers__.append(module)
    init_sqlite(
        DRIVERS[module],
        getenv(
            'DBEXPLORER_CFG',
            expanduser('~/.dbexplorer/dbexplorer.cfg')),
        getenv(
            'SQLITEPASS_CFG',
            expanduser('~/.sqlitepass')),
        getenv(
            'NAVICAT_CFG',
            expanduser('~/Library/Application Support/PremiumSoft CyberTech'
                       '/preference.plist')),
        getenv(
            'NAVICAT_CFG',
            expanduser('~/Library/Containers/com.prect.'
                       'NavicatEssentialsForSQLite/Data/Library/'
                       'Application Support/PremiumSoft CyberTech/'
                       'preference.plist'))
    )

    Options.parser['sqlite'] = SQLiteOptionsParser()



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.writer import FormatWriter
from dbmanagr.formatter import Formatter, AutocompleteFormatter
from dbmanagr.formatter import JsonFormatter, SimpleFormatter, \
    SimplifiedFormatter


class SimplifiedWriter(FormatWriter):
    def __init__(self):
        FormatWriter.__init__(self, u'{0}\n')
        Formatter.set(SimplifiedFormatter())


class JsonWriter(FormatWriter):
    def __init__(self):
        FormatWriter.__init__(
            self,
            u"""[
{0}
]
""", item_separator=u',\n',)
        Formatter.set(JsonFormatter())


class AutocompleteWriter(FormatWriter):
    def __init__(self):
        FormatWriter.__init__(self, u'{0}', u'{autocomplete}')
        Formatter.set(AutocompleteFormatter())


class SimpleWriter(FormatWriter):
    def __init__(self):
        FormatWriter.__init__(
            self,
            u"""Id\tTitle\tSubtitle\tAutocomplete
{0}
""")
        Formatter.set(SimpleFormatter())



# -*- coding: utf-8 -*-
#
# Copyright © 2014 René Samselnig
#
# This file is part of Database Navigator.
#
# Database Navigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Database Navigator is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Database Navigator.  If not, see <http://www.gnu.org/licenses/>.
#

from dbmanagr.dto import Dto
from dbmanagr.jsonable import from_json
from dbmanagr.utils import filter_keys


class Table(Dto):
    def __init__(
            self,
            name=None,
            uri=None,
            owner=None,
            size=None,
            primary_key=None,
            columns=None,
            foreign_keys=None,
            autocomplete=None):
        Dto.__init__(self, autocomplete)

        self.name = name
        self.uri = uri
        self.owner = owner
        self.size = size
        self.primary_key = primary_key
        self.columns = columns
        self.foreign_keys = foreign_keys

    def __str__(self):
        if self.uri:
            return '{}{}'.format(self.uri, self.name)
        return self.name

    @staticmethod
    def from_json(d):
        return Table(
            **from_json(
                filter_keys(
                    d,
                    'name', 'uri', 'owner', 'size', 'primary_key', 'columns',
                    'foreign_keys', 'autocomplete'
                )
            )
        )
